@@ -3,9 +3,9 @@ package App::ClusterSSH;
 use 5.008.004;
 use warnings;
 use strict;
-use version; our $VERSION = version->new('4.03_06');
+use version; our $VERSION = version->new('4.10_01');
 
-use Carp qw/cluck/;
+use Carp qw/cluck :DEFAULT/;
 
 use base qw/ App::ClusterSSH::Base /;
 use App::ClusterSSH::Host;
@@ -27,16 +27,18 @@ require Tk::Dialog;
 require Tk::LabEntry;
 use X11::Protocol;
 use X11::Protocol::Constants qw/ Shift Mod5 ShiftMask /;
+use X11::Protocol::WM;
 use vars qw/ %keysymtocode %keycodetosym /;
 use X11::Keysyms '%keysymtocode', 'MISCELLANY', 'XKB_KEYS', '3270', 'LATIN1',
     'LATIN2', 'LATIN3', 'LATIN4', 'KATAKANA', 'ARABIC', 'CYRILLIC', 'GREEK',
     'TECHNICAL', 'SPECIAL', 'PUBLISHING', 'APL', 'HEBREW', 'THAI', 'KOREAN';
 use File::Basename;
+use Module::Load;
 use Net::hostent;
-use Carp;
 use Sys::Hostname;
 use English;
 use Socket;
+use File::Path qw(make_path);
 
 # Notes on general order of processing
 #
@@ -59,6 +61,8 @@ sub new {
     $self->{config}  = App::ClusterSSH::Config->new( parent => $self, );
     $self->{helper}  = App::ClusterSSH::Helper->new( parent => $self, );
 
+    $self->set_config( $self->config );
+
     # catch and reap any zombies
     $SIG{CHLD} = sub {
         my $kid;
@@ -109,6 +113,12 @@ my $xdisplay;
 my %keyboardmap;
 my $sysconfigdir = "/etc";
 my %ssh_hostnames;
+my $host_menu_static_items;    # number of items in the host menu that should
+                               # not be touched by build_host_menu
+my (@dead_hosts);              # list of hosts whose sessions are now closed
+my $sort = sub { sort @_ };   # reference to our sort function which may later
+                              # be changed in run() if the user has asked for
+                              # natural sorting
 
 $keysymtocode{unknown_sym} = 0xFFFFFF;    # put in a default "unknown" entry
 $keysymtocode{EuroSign}
@@ -504,6 +514,39 @@ SWITCH: {
     return $self;
 }
 
+sub substitute_macros {
+    my ( $self, $svr, $text ) = @_;
+
+    return $text unless ( $self->config->{macros_enabled} eq 'yes' );
+
+    {
+        my $macro_servername = $self->config->{macro_servername};
+        ( my $servername = $svr ) =~ s/\s+//;
+        $text =~ s!$macro_servername!$servername!xsmg;
+    }
+    {
+        my $macro_hostname = $self->config->{macro_hostname};
+        my $hostname       = $servers{$svr}{givenname};
+        $text =~ s!$macro_hostname!$hostname!xsmg;
+    }
+    {
+        my $macro_username = $self->config->{macro_username};
+        my $username       = $servers{$svr}{username};
+        $username ||= getpwuid($UID);
+        $text =~ s!$macro_username!$username!xsmg;
+    }
+    {
+        my $macro_newline = $self->config->{macro_newline};
+        $text =~ s!$macro_newline!\n!xsmg;
+    }
+    {
+        my $macro_version = $self->config->{macro_version};
+        $text =~ s/$macro_version/$VERSION/xsmg;
+    }
+
+    return $text;
+}
+
 sub send_text($@) {
     my $self = shift;
     my $svr  = shift;
@@ -512,34 +555,7 @@ sub send_text($@) {
     $self->debug( 2, "servers{$svr}{wid}=$servers{$svr}{wid}" );
     $self->debug( 3, "Sending to '$svr' text:$text:" );
 
-    # command macro substitution
-    if ( $self->config->{macros_enabled} eq 'yes' ) {
-
-        # $svr contains a trailing space here, so ensure its stripped off
-        {
-            my $macro_servername = $self->config->{macro_servername};
-            my $servername       = $svr;
-            $servername =~ s/\s+//;
-            $text =~ s/$macro_servername/$servername/xsmg;
-        }
-        $text =~ s/%h/hostname()/xsmeg;
-
-        # use connection username, else default to current username
-        {
-            my $macro_username = $self->config->{macro_username};
-            my $username       = $servers{$svr}{username};
-            $username ||= getpwuid($UID);
-            $text =~ s/$macro_username/$username/xsmg;
-        }
-        {
-            my $macro_newline = $self->config->{macro_newline};
-            $text =~ s/$macro_newline/\n/xsmg;
-        }
-        {
-            my $macro_version = $self->config->{macro_version};
-            $text =~ s/$macro_version/$VERSION/xsmg;
-        }
-    }
+    $text = $self->substitute_macros( $svr, $text );
 
     foreach my $char ( split( //, $text ) ) {
         next if ( !defined($char) );
@@ -572,7 +588,6 @@ sub send_text($@) {
                     'name'        => $event,
                     'detail'      => $code,
                     'state'       => $state,
-                    'time'        => time(),
                     'event'       => $servers{$svr}{wid},
                     'root'        => $xdisplay->root(),
                     'same_screen' => 1,
@@ -718,6 +733,12 @@ sub open_client_windows(@) {
           # Since this is the child, we can mark any server unresolved without
           # affecting the main program
             $servers{$server}{realname} .= "==" if ( !$realname );
+
+            # copy and amend the config provided to the helper script
+            my $local_config = $self->config;
+            $local_config->{command} = $self->substitute_macros( $server,
+                $local_config->{command} );
+
             my $exec = join( ' ',
                 $self->config->{terminal},
                 $color,
@@ -867,6 +888,17 @@ sub show_console() {
     return $self;
 }
 
+# set the first argument to the second if the first is undefined
+# the equivalent of //= but works in older Perls (e.g. 5.8)
+sub slash_slash_equal(\$$) {
+
+    if ( !defined( ${ $_[0] } ) ) {
+        ${ $_[0] } = $_[1];
+    }
+
+    return ${ $_[0] };
+}
+
 # leave function def open here so we can be flexible in how it's called
 sub retile_hosts {
     my ( $self, $force ) = @_;
@@ -964,14 +996,50 @@ sub retile_hosts {
         );
     }
 
-    $self->config->dump("noexit") if ( $self->options->debug > 1 );
+    $self->config->dump("noexit") if ( $self->options->debug_level > 1 );
+
+    # now find the size of the window decorations
+    if ( !exists( $self->config->{internal_terminal_wm_decoration_left} ) ) {
+
+   # Debian #842965 (https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=842965)
+   # disable behavior added in https://github.com/duncs/clusterssh/pull/66
+   # unless explicitly enabled with auto_wm_decoration_offsets => yes
+
+        if ( $self->config->{auto_wm_decoration_offsets} =~ /yes/i ) {
+
+            # use the first window as exemplary
+            my ($wid) = $servers{ ( keys(%servers) )[0] }{wid};
+
+            if ( defined($wid) ) {
+
+                # get the WM decoration sizes
+                (   $self->config->{internal_terminal_wm_decoration_left},
+                    $self->config->{internal_terminal_wm_decoration_right},
+                    $self->config->{internal_terminal_wm_decoration_top},
+                    $self->config->{internal_terminal_wm_decoration_bottom}
+                    )
+                    = X11::Protocol::WM::get_net_frame_extents( $xdisplay,
+                    $wid );
+            }
+        }
+
+        # in case the WM call failed we set some defaults
+        slash_slash_equal(
+            $self->config->{internal_terminal_wm_decoration_left}, 0 );
+        slash_slash_equal(
+            $self->config->{internal_terminal_wm_decoration_right}, 0 );
+        slash_slash_equal(
+            $self->config->{internal_terminal_wm_decoration_top}, 0 );
+        slash_slash_equal(
+            $self->config->{internal_terminal_wm_decoration_bottom}, 0 );
+    }
 
     # now we have the info, plot first window position
     my @hosts;
     my ( $current_x, $current_y, $current_row, $current_col ) = 0;
     if ( $self->config->{window_tiling_direction} =~ /right/i ) {
         $self->debug( 2, "Tiling top left going bot right" );
-        @hosts     = sort( keys(%servers) );
+        @hosts     = $sort->( keys(%servers) );
         $current_x = $self->config->{screen_reserve_left}
             + $self->config->{terminal_reserve_left};
         $current_y = $self->config->{screen_reserve_top}
@@ -981,7 +1049,7 @@ sub retile_hosts {
     }
     else {
         $self->debug( 2, "Tiling bot right going top left" );
-        @hosts = reverse( sort( keys(%servers) ) );
+        @hosts = reverse( $sort->( keys(%servers) ) );
         $current_x
             = $self->config->{screen_reserve_right}
             - $self->config->{internal_screen_width}
@@ -1029,14 +1097,18 @@ sub retile_hosts {
             $current_x
                 += $self->config->{terminal_reserve_left}
                 + $self->config->{terminal_reserve_right}
-                + $self->config->{internal_terminal_width};
+                + $self->config->{internal_terminal_width}
+                + $self->config->{internal_terminal_wm_decoration_left}
+                + $self->config->{internal_terminal_wm_decoration_right};
 
             $current_col += 1;
             if ( $current_col == $self->config->{internal_columns} ) {
                 $current_y
                     += $self->config->{terminal_reserve_top}
                     + $self->config->{terminal_reserve_bottom}
-                    + $self->config->{internal_terminal_height};
+                    + $self->config->{internal_terminal_height}
+                    + $self->config->{internal_terminal_wm_decoration_top}
+                    + $self->config->{internal_terminal_wm_decoration_bottom};
                 $current_x = $self->config->{screen_reserve_left}
                     + $self->config->{terminal_reserve_left};
                 $current_row++;
@@ -1085,7 +1157,7 @@ sub capture_terminal() {
     my ($self) = @_;
     $self->debug( 0, "Stub for capturing a terminal window" );
 
-    return if ( $self->coptions->debug < 6 );
+    return if ( $self->options->debug_level < 6 );
 
     # should never see this - all experimental anyhow
 
@@ -1231,13 +1303,43 @@ sub add_host_by_name() {
     }
 }
 
+# attempt to re-add any hosts that have been closed since we started
+# the session - either through errors or deliberate log-outs
+sub re_add_closed_sessions() {
+    my ($self) = @_;
+    $self->debug( 2, "add closed sessions" );
+
+    return if ( scalar(@dead_hosts) == 0 );
+
+    my @new_hosts = @dead_hosts;
+
+    # clear out the list in case open fails
+    @dead_hosts = qw//;
+
+    # try to open
+    $self->open_client_windows(@new_hosts);
+
+    # update hosts list with current state
+    $self->build_hosts_menu();
+
+    # retile, or bring console to front
+    if ( $self->config->{window_tiling} eq "yes" ) {
+        return $self->retile_hosts();
+    }
+    else {
+        return $self->show_console();
+    }
+}
+
 sub build_hosts_menu() {
     my ($self) = @_;
+
+    return if ( $self->config->{hide_menu} );
+
     $self->debug( 2, "Building hosts menu" );
 
-    # first, empty the hosts menu from the 4th entry on
+    # first, empty the hosts menu from the last static entry + 1 on
     my $menu = $menus{bar}->entrycget( 'Hosts', -menu );
-    my $host_menu_static_items = 7;
     $menu->delete( $host_menu_static_items, 'end' );
 
     $self->debug( 3, "Menu deleted" );
@@ -1248,7 +1350,7 @@ sub build_hosts_menu() {
     $self->debug( 3, "Parsing list" );
 
     my $menu_item_counter = $host_menu_static_items;
-    foreach my $svr ( sort( keys(%servers) ) ) {
+    foreach my $svr ( $sort->( keys(%servers) ) ) {
         $self->debug( 3, "Checking $svr and restoring active value" );
         my $colbreak = 0;
         if ( $menu_item_counter > $self->config->{max_host_menu_items} ) {
@@ -1293,6 +1395,7 @@ sub setup_repeat() {
                 if ( defined( $servers{$svr}{pid} ) ) {
                     if ( !kill( 0, $servers{$svr}{pid} ) ) {
                         $build_menu = 1;
+                        push( @dead_hosts, $servers{$svr}{connect_string} );
                         delete( $servers{$svr} );
                         $self->debug( 0, "$svr session closed" );
                     }
@@ -1647,11 +1750,14 @@ sub key_event {
                 $self->debug( 3, "matched combo" );
                 if ( $event eq "KeyRelease" ) {
                     $self->debug( 2, "Received hotkey: $hotkey" );
-                    $self->send_text_to_all_servers('%s')
+                    $self->send_text_to_all_servers(
+                        $self->config->{macro_servername} )
                         if ( $hotkey eq "key_clientname" );
-                    $self->send_text_to_all_servers('%h')
+                    $self->send_text_to_all_servers(
+                        $self->config->{macro_hostname} )
                         if ( $hotkey eq "key_localname" );
-                    $self->send_text_to_all_servers('%u')
+                    $self->send_text_to_all_servers(
+                        $self->config->{macro_username} )
                         if ( $hotkey eq "key_username" );
                     $self->add_host_by_name()
                         if ( $hotkey eq "key_addhost" );
@@ -1689,7 +1795,6 @@ sub key_event {
                     'name'        => $event,
                     'detail'      => $keycode,
                     'state'       => $state,
-                    'time'        => time(),
                     'event'       => $servers{$_}{wid},
                     'root'        => $xdisplay->root(),
                     'same_screen' => 1,
@@ -1706,7 +1811,9 @@ sub create_menubar() {
     my ($self) = @_;
     $self->debug( 2, "create_menubar: started" );
     $menus{bar} = $windows{main_window}->Menu();
-    $windows{main_window}->configure( -menu => $menus{bar}, );
+
+    $windows{main_window}->configure( -menu => $menus{bar}, )
+        unless $self->config->{hide_menu};
 
     $menus{file} = $menus{bar}->cascade(
         -label     => 'File',
@@ -1725,42 +1832,51 @@ sub create_menubar() {
         -tearoff => 0,
     );
 
+    my $host_menu_items = [
+        [   "command",
+            "Retile Windows",
+            -command => sub { $self->retile_hosts },
+            -accelerator => $self->config->{key_retilehosts},
+        ],
+
+#         [ "command", "Capture Terminal",    -command => sub { $self->capture_terminal), ],
+        [   "command",
+            "Set all active",
+            -command => sub { $self->set_all_active() },
+        ],
+        [   "command",
+            "Set half inactive",
+            -command => sub { $self->set_half_inactive() },
+        ],
+        [   "command",
+            "Toggle active state",
+            -command => sub { $self->toggle_active_state() },
+        ],
+        [   "command",
+            "Close inactive sessions",
+            -command => sub { $self->close_inactive_sessions() },
+        ],
+        [   "command",
+            "Add Host(s) or Cluster(s)",
+            -command => sub { $self->add_host_by_name, },
+            -accelerator => $self->config->{key_addhost},
+        ],
+        [   "command",
+            "Re-add closed session(s)",
+            -command => sub { $self->re_add_closed_sessions() },
+        ],
+        ''      # this is needed as build_host_menu always drops the
+                # last item
+    ];
+
     $menus{hosts} = $menus{bar}->cascade(
         -label     => 'Hosts',
         -tearoff   => 1,
-        -menuitems => [
-            [   "command",
-                "Retile Windows",
-                -command => sub { $self->retile_hosts },
-                -accelerator => $self->config->{key_retilehosts},
-            ],
-
-#         [ "command", "Capture Terminal",    -command => sub { $self->capture_terminal), ],
-            [   "command",
-                "Set all active",
-                -command => sub { $self->set_all_active() },
-            ],
-            [   "command",
-                "Set half inactive",
-                -command => sub { $self->set_half_inactive() },
-            ],
-            [   "command",
-                "Toggle active state",
-                -command => sub { $self->toggle_active_state() },
-            ],
-            [   "command",
-                "Close inactive sessions",
-                -command => sub { $self->close_inactive_sessions() },
-            ],
-            [   "command",
-                "Add Host(s) or Cluster(s)",
-                -command => sub { $self->add_host_by_name, },
-                -accelerator => $self->config->{key_addhost},
-            ],
-            '',
-        ],
+        -menuitems => $host_menu_items
     );
 
+    $host_menu_static_items = scalar( @{$host_menu_items} );
+
     $menus{send} = $menus{bar}->cascade(
         -label   => 'Send',
         -tearoff => 1,
@@ -1932,16 +2048,19 @@ sub run {
     # only use ssh_args from options if config file ssh_args not set AND
     # options is not the default value otherwise the default options
     # value is used instead of the config file
-    if ( $self->config->{ssh_args} ) {
-        if (   $self->options->options
-            && $self->options->options ne $self->options->options_default )
-        {
-            $self->config->{ssh_args} = $self->options->options;
+    if ( $self->config->{comms} eq 'ssh' ) {
+        if ( $self->config->{ssh_args} ) {
+            if (   $self->options->options
+                && $self->options->options ne
+                $self->options->options_default )
+            {
+                $self->config->{ssh_args} = $self->options->options;
+            }
+        }
+        else {
+            $self->config->{ssh_args} = $self->options->options
+                if ( $self->options->options );
         }
-    }
-    else {
-        $self->config->{ssh_args} = $self->options->options
-            if ( $self->options->options );
     }
 
     $self->config->{terminal_args} = $self->options->term_args
@@ -1952,6 +2071,20 @@ sub run {
             = "-xrm '$1.VT100.allowSendEvents:true'";
     }
 
+    # if the user has asked for natural sorting we need to include an extra
+    # module
+    if ( $self->config()->{'use_natural_sort'} ) {
+        eval { Module::Load::load('Sort::Naturally'); };
+        if ($@) {
+            warn(
+                "natural sorting requested but unable to load Sort::Naturally: $@\n"
+            );
+        }
+        else {
+            $sort = sub { Sort::Naturally::nsort(@_) };
+        }
+    }
+
     $self->config->dump() if ( $self->options->dump_config );
 
     $self->evaluate_commands() if ( $self->options->evaluate );
@@ -1971,14 +2104,32 @@ sub run {
     $self->cluster->get_tag_entries( split /,/,
         $self->config->{extra_tag_file} || '' );
 
-    if ( $self->options->list ) {
-        print( 'Available cluster tags:', $/ );
-        print "\t", $_, $/ foreach ( sort( $self->cluster->list_tags ) );
+    if ( defined $self->options->list ) {
+        my $eol = $self->options->quiet ? ' ' : $/;
+        my $tab = $self->options->quiet ? ''  : "\t";
+        if ( !$self->options->list ) {
+            print( 'Available cluster tags:', $/ )
+                unless ( $self->options->quiet );
+            print $tab, $_, $eol
+                foreach ( sort( $self->cluster->list_tags ) );
 
-        my @external_clusters = $self->cluster->list_external_clusters;
-        if (@external_clusters) {
-            print( 'Available external command tags:', $/ );
-            print "\t", $_, $/ foreach ( sort(@external_clusters) );
+            my @external_clusters = $self->cluster->list_external_clusters;
+            if (@external_clusters) {
+                print( 'Available external command tags:', $/ )
+                    unless ( $self->options->quiet );
+                print $tab, $_, $eol foreach ( sort(@external_clusters) );
+                print $/;
+            }
+        }
+        else {
+            print 'Tag resolved to hosts: ', $/
+                unless ( $self->options->quiet );
+            @servers = $self->resolve_names( $self->options->list );
+
+            foreach my $svr (@servers) {
+                print $tab, $svr, $eol;
+            }
+            print $/;
         }
 
         $self->debug(
@@ -2011,7 +2162,8 @@ sub run {
     $self->debug( 2, "Capture map events" );
     $self->capture_map_events();
 
-    $self->debug( 0, 'Opening to: ', join( ' ', @servers ) ) if (@servers);
+    $self->debug( 0, 'Opening to: ', join( ' ', @servers ) )
+        if ( @servers && !$self->options->quiet );
     $self->open_client_windows(@servers);
 
     # Check here if we are tiling windows.  Here instead of in func so
@@ -2138,10 +2290,17 @@ the code until this time.
 
 =item populate_send_menu_entries_from_xml
 
+=item re_add_closed_sessions
+
 =item remove_repeated_servers
 
 =item resolve_names
 
+=item slash_slash_equal
+
+An implementation of the //= operator that works on older Perls.
+slash_slash_equal($a, 0) is equivalent to $a //= 0
+
 =item retile_hosts
 
 =item run
@@ -2164,6 +2323,8 @@ the code until this time.
 
 =item show_history
 
+=item substitute_macros
+
 =item terminate_host
 
 =item toggle_active_state
@@ -2218,7 +2379,7 @@ Duncan Ferguson, C<< <duncan_j_ferguson at yahoo.co.uk> >>
 
 =head1 COPYRIGHT & LICENSE
 
-Copyright 1999-2010 Duncan Ferguson, all rights reserved.
+Copyright 1999-2016 Duncan Ferguson, all rights reserved.
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of either: the GNU General Public License as published
