@@ -37,30 +37,44 @@ sub new {
 
     # load in ssh hostname for later use
     if ( !%ssh_hostname_for || !$ssh_configs_read{ $self->{ssh_config} } ) {
-        $ssh_configs_read{ $self->{ssh_config} } = 1;
-        if ( open( my $ssh_config_fh, '<', $self->{ssh_config} ) ) {
-            while ( my $line = <$ssh_config_fh> ) {
-                chomp $line;
-                next unless ( $line =~ m/^\s*host\s+(.*)/i );
+        $self->read_ssh_file( $self->{ssh_config} );
 
-                # account for multiple declarations of hosts
-                $ssh_hostname_for{$_} = 1 foreach ( split( /\s+/, $1 ) );
-            }
-            close($ssh_config_fh);
-
-            $self->debug( 5, 'Have the following ssh hostnames' );
-            $self->debug( 5, '  "', $_, '"' )
-                foreach ( sort keys %ssh_hostname_for );
-        }
-        else {
-            $self->debug( 3, 'Unable to read ',
-                $self->{ssh_config}, ': ', $!, $/ );
-        }
+        $self->debug( 5, 'Have the following ssh hostnames' );
+        $self->debug( 5, '  "', $_, '"' )
+            foreach ( sort keys %ssh_hostname_for );
     }
 
     return $self;
 }
 
+sub read_ssh_file($$) {
+    my ($self)     = shift;
+    my ($filename) = glob(shift);
+    $self->debug( 3, 'Reading SSH file: ', $filename );
+
+    $ssh_configs_read{$filename} = 1;
+
+    if ( open( my $ssh_config_fh, '<', $filename ) ) {
+        while ( my $line = <$ssh_config_fh> ) {
+            chomp $line;
+
+            if ( $line =~ /^\s*include\s+(.+)/i ) {
+                $self->read_ssh_file($1);
+                next;
+            }
+
+            next unless ( $line =~ m/^\s*host\s+(.*)/i );
+
+            # account for multiple declarations of hosts
+            $ssh_hostname_for{$_} = 1 foreach ( split( /\s+/, $1 ) );
+        }
+        close($ssh_config_fh);
+    }
+    else {
+        $self->debug( 3, 'Unable to read ', $filename, ': ', $!, $/ );
+    }
+}
+
 sub get_hostname {
     my ($self) = @_;
     return $self->{hostname};
@@ -73,7 +87,7 @@ sub get_username {
 
 sub get_type {
     my ($self) = @_;
-    if($self->check_ssh_hostname) {
+    if ( $self->check_ssh_hostname ) {
         return 'ssh_alias';
     }
     return $self->{type} || q{};
@@ -140,9 +154,6 @@ sub get_realname {
                 : $self->{hostname};
         }
     }
-    else {
-        $self->{realname} = $self->{hostname};
-    }
     return $self->{realname};
 }
 
@@ -165,7 +176,10 @@ sub parse_host_string {
     {
         $self->debug(
             5,
-            $self->loc( 'bracketed IPv6: u=[_1] h=[_2] p=[_3] g=[_4]', $1, $2, $3, $4 ),
+            $self->loc(
+                'bracketed IPv6: u=[_1] h=[_2] p=[_3] g=[_4]',
+                $1, $2, $3, $4
+            ),
         );
         return __PACKAGE__->new(
             parse_string => $parse_string,
@@ -188,8 +202,12 @@ sub parse_host_string {
         }xms
         )
     {
-        $self->debug( 5,
-            $self->loc( 'std IPv4: u=[_1] h=[_2] p=[_3] g=[_4]', $1, $2, $3, $4 ),
+        $self->debug(
+            5,
+            $self->loc(
+                'std IPv4: u=[_1] h=[_2] p=[_3] g=[_4]',
+                $1, $2, $3, $4
+            ),
         );
         return __PACKAGE__->new(
             parse_string => $parse_string,
@@ -218,9 +236,9 @@ sub parse_host_string {
         $geometry = $1;
     }
 
-    # Check for a '/nnnn' port definition 
+    # Check for a '/nnnn' port definition
     if ( $host_string =~ s!(?:/(\d+)$)!! ) {
-       $port = $1;
+        $port = $1;
     }
 
     # use number of colons as a possible indicator
@@ -230,7 +248,7 @@ sub parse_host_string {
     # if its 8 then assumed full IPv6 address with a port
     # also catch localhost address here
     if ( $colon_count == 7 || $colon_count == 8 || $host_string eq '::1' ) {
-        if( $colon_count == 8) {
+        if ( $colon_count == 8 ) {
             $host_string =~ s/(?::(\d+?))$//;
             $port = $1;
         }
@@ -251,8 +269,7 @@ sub parse_host_string {
         );
     }
     if (   $colon_count > 1
-        && $colon_count < 8
-    )
+        && $colon_count < 8 )
     {
         warn 'Ambiguous host string: "', $host_string, '"',   $/;
         warn 'Assuming you meant "[',    $host_string, ']"?', $/;
@@ -260,8 +277,8 @@ sub parse_host_string {
         $self->debug(
             5,
             $self->loc(
-                'Ambiguous IPv6 u=[_1] h=[_2] p=[_3] g=[_4]', $username,
-                $host_string, $port, $geometry,
+                'Ambiguous IPv6 u=[_1] h=[_2] p=[_3] g=[_4]',
+                $username, $host_string, $port, $geometry,
             )
         );
 
@@ -383,6 +400,11 @@ Given a host string, returns a host object.  Parses hosts such as
 Check the objects hostname to see whether or not it may be configured within 
 the users F< $HOME/.ssh/config > configuration file
 
+=item read_ssh_file
+
+Method to ease reading in ssh configuration files.  Used for grabbing
+hostnames for validation when used in clusters
+
 =over 4
 
 =item host
@@ -414,7 +436,7 @@ Duncan Ferguson, C<< <duncan_j_ferguson at yahoo.co.uk> >>
 
 =head1 LICENSE AND COPYRIGHT
 
-Copyright 1999-2010 Duncan Ferguson.
+Copyright 1999-2016 Duncan Ferguson.
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of either: the GNU General Public License as published
